from abc import ABCMeta, abstractmethod
from itertools import islice
from operator import attrgetter
import re
from typing import Iterable, TYPE_CHECKING  # noqa: F401

from six import add_metaclass
from six.moves import filter


if TYPE_CHECKING:
    from wsgi_lineprof.stats import LineProfilerStat  # noqa: F401


@add_metaclass(ABCMeta)
class BaseFilter(object):
    @abstractmethod
    def filter(self, stats):
        # type: (Iterable[LineProfilerStat]) -> Iterable[LineProfilerStat]
        pass


class FilenameFilter(BaseFilter):
    """Filter which matches with filename"""
    def __init__(self, filename, regex=False):
        # type: (str, bool) -> None
        self.filename = filename
        self.regex = regex
        self.compiled_regex = re.compile(filename)

    def filter(self, stats):
        # type: (Iterable[LineProfilerStat]) -> Iterable[LineProfilerStat]
        if self.regex:
            compiled_regex = self.compiled_regex
            return filter(lambda s: compiled_regex.search(s.filename), stats)
        else:
            return filter(lambda s: self.filename in s.filename, stats)


class NameFilter(BaseFilter):
    """Filter which matches with name"""
    def __init__(self, name, regex=True):
        # type: (str, bool) -> None
        self.name = name
        self.regex = regex
        self.compiled_regex = re.compile(name)

    def filter(self, stats):
        # type: (Iterable[LineProfilerStat]) -> Iterable[LineProfilerStat]
        if self.regex:
            compiled_regex = self.compiled_regex
            return filter(lambda s: compiled_regex.search(s.name), stats)
        else:
            return filter(lambda s: self.name in s.name, stats)


class TotalTimeSorter(BaseFilter):
    """Sort stats by total time"""
    def __init__(self, reverse=True):
        # type: (bool) -> None
        self.reverse = reverse

    def filter(self, stats):
        # type: (Iterable[LineProfilerStat]) -> Iterable[LineProfilerStat]
        return sorted(stats,
                      key=attrgetter("total_time"), reverse=self.reverse)


class TopItemsFilter(BaseFilter):
    """Get first n stats"""
    def __init__(self, n=10):
        # type: (int) -> None
        self.n = n

    def filter(self, stats):
        # type: (Iterable[LineProfilerStat]) -> Iterable[LineProfilerStat]
        return islice(stats, self.n)



from cpython cimport PyObject


cdef extern from "Python.h":
    ctypedef struct PyCodeObject:
        int       co_argcount
        int       co_nlocals
        int       co_stacksize
        int       co_flags
        PyObject *co_code
        PyObject *co_consts
        PyObject *co_names
        PyObject *co_varnames
        PyObject *co_freevars
        PyObject *co_cellvars
        PyObject *co_filename
        PyObject *co_name
        int       co_firstlineno
        PyObject *co_lnotab

    ctypedef struct PyFrameObject:
        PyFrameObject *f_back
        PyCodeObject  *f_code
        PyObject *f_builtins
        PyObject *f_globals
        PyObject *f_locals
        PyObject *f_trace
        PyObject *f_exc_type
        PyObject *f_exc_value
        PyObject *f_exc_traceback
        int f_lasti
        int f_lineno
        int f_restricted
        int f_iblock
        int f_nlocals
        int f_ncells
        int f_nfreevars
        int f_stacksize

cdef extern from "frameobject.h":
    ctypedef int (*Py_tracefunc)(object self, PyFrameObject *py_frame, int what, PyObject *arg)

cdef extern from "Python.h":
    ctypedef long long PY_LONG_LONG
    cdef bint PyCFunction_Check(object obj)

    cdef void PyEval_SetProfile(Py_tracefunc func, object arg)
    cdef void PyEval_SetTrace(Py_tracefunc func, object arg)

    ctypedef object (*PyCFunction)(object self, object args)

    ctypedef struct PyMethodDef:
        char *ml_name
        PyCFunction ml_meth
        int ml_flags
        char *ml_doc

    ctypedef struct PyCFunctionObject:
        PyMethodDef *m_ml
        PyObject *m_self
        PyObject *m_module

    cdef int PyTrace_LINE
    cdef int PyTrace_RETURN



from __future__ import print_function
import re
import sys
from wsgiref.simple_server import make_server
from wsgiref.util import setup_testing_defaults

import jinja2
from six import StringIO

from wsgi_lineprof.middleware import LineProfilerMiddleware


RE_IP = (
    r"(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}"
    r"(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)"
)
RE_EMAIL = (
    r"([-!#-'*+/-9=?A-Z^-~]+(\.[-!#-'*+/-9=?A-Z^-~]+)*"
    r"|\"([]!#-[^-~ \t]|(\\[\t -~]))+\")@[0-9A-Za-z]"
    r"([0-9A-Za-z-]{0,61}[0-9A-Za-z])?"
    r"(\.[0-9A-Za-z]([0-9A-Za-z-]{0,61}[0-9A-Za-z])?)+"
)
JINJA_TEMPLATE = """<!DOCTYPE html>
<html lang="en">
<head>
  <title>{{ title }}</title>
</head>
<body>
  <h1>{{ title }}</h1>
  <ul>
  {% for user in users %}
    <li>
      <a href="/users/{{ user.url }}">{{ user.name }}</a>
      {% if loop.index is divisibleby 2 %}Active{% else %}Inactive{% endif %}
    </li>
  {% endfor %}
  </ul>
</body>
</html>"""


def demo_app(environ, start_response):
    """Copy of wsgiref.simple_server.demo_app"""
    stdout = StringIO()
    print("Hello world!", file=stdout)
    print(file=stdout)
    h = sorted(environ.items())
    for k, v in h:
        print(k, '=', repr(v), file=stdout)
    start_response("200 OK", [('Content-Type', 'text/plain; charset=utf-8')])
    return [stdout.getvalue().encode("utf-8")]


def fib(n):
    if n <= 1:
        return n
    return fib(n - 1) + fib(n - 2)


def fib_app(environ, start_response):
    setup_testing_defaults(environ)

    status = '200 OK'
    headers = [('Content-type', 'text/plain; charset=utf-8')]

    start_response(status, headers)

    m = re.match(r"/([1-9][0-9]*)", environ["PATH_INFO"])
    if m:
        n = int(m.group(1))
    else:
        n = 5

    fib_n = fib(n)
    res = "fib(%d) = %d" % (n, fib_n)
    return [res.encode("utf-8")]


def re_app(environ, start_response):
    setup_testing_defaults(environ)

    status = '200 OK'
    headers = [('Content-type', 'text/plain; charset=utf-8')]

    start_response(status, headers)

    text = environ["PATH_INFO"]

    res = ""

    re.match(RE_EMAIL, text)
    res += str(re.search(RE_EMAIL, text))
    res += "\n"

    re.match(RE_IP, text)
    res += str(re.search(RE_IP, text))
    res += "\n"

    return [res.encode("utf-8")]


def jinja_app(environ, start_response):
    setup_testing_defaults(environ)

    status = '200 OK'
    headers = [('Content-type', 'text/html; charset=utf-8')]

    start_response(status, headers)

    m = re.match(r"/([1-9][0-9]*)", environ["PATH_INFO"])
    if m:
        n = int(m.group(1))
    else:
        n = 5

    template = jinja2.Template(JINJA_TEMPLATE)
    users = [{"url": str(i), "name": "A" * i} for i in range(n)]
    res = template.render(title="Benchmark", users=users)
    return [res.encode("utf-8")]


if __name__ == "__main__":
    if len(sys.argv) <= 1:
        sys.exit(1)

    app_name = sys.argv[1]
    if not app_name.endswith("_app"):
        sys.exit(1)
    app = locals()[app_name]
    # Set up profiler
    app = LineProfilerMiddleware(app)

    server = make_server('127.0.0.1', 8000, app)
    print("Serving on 127.0.0.1:8000...")
    server.serve_forever()



import random
import string

from webtest import TestApp

from wsgi_lineprof.filters import FilenameFilter
from wsgi_lineprof.middleware import LineProfilerMiddleware
from .apps import demo_app, fib_app, jinja_app, re_app


class StringNoopIO(object):
    def write(self, _):
        pass


def prepare_app(app, profiler, filters=None):
    if filters is None:
        filters = []

    if profiler == "sync":
        app = LineProfilerMiddleware(app,
                                     filters=filters,
                                     stream=StringNoopIO())
    elif profiler == "async":
        app = LineProfilerMiddleware(app,
                                     filters=filters,
                                     stream=StringNoopIO(),
                                     async_stream=True)
    return TestApp(app)


class DemoAppTest(object):
    param_names = ["profiler"]
    params = ["base", "sync", "async"]

    def setup(self, profiler):
        self.app = prepare_app(demo_app, profiler)

    def time_index_page(self, *args):
        self.app.get("/")


class FibAppTest(object):
    param_names = ["profiler"]
    params = ["base", "sync", "async"]

    def setup(self, profiler):
        self.app = prepare_app(fib_app, profiler)

    def time_index_page(self, *args):
        self.app.get("/")

    def time_n_10(self, *args):
        self.app.get("/10")

    def time_n_20(self, *args):
        self.app.get("/20")


class FibAppWithFilenameFilterTest(object):
    param_names = ["profiler"]
    params = ["base", "sync", "async"]

    def setup(self, profiler):
        filters = [
            FilenameFilter("apps.py"),
        ]
        self.app = prepare_app(fib_app, profiler, filters=filters)

    def time_index_page(self, *args):
        self.app.get("/")

    def time_n_10(self, *args):
        self.app.get("/10")

    def time_n_20(self, *args):
        self.app.get("/20")


class JinjaAppTest(object):
    param_names = ["profiler"]
    params = ["base", "sync", "async"]

    def setup(self, profiler):
        self.app = prepare_app(jinja_app, profiler)

    def time_10_items(self, *args):
        self.app.get("/10")

    def time_100_items(self, *args):
        self.app.get("/20")


class JinjaAppWithFilenameFilterTest(object):
    param_names = ["profiler"]
    params = ["base", "sync", "async"]

    def setup(self, profiler):
        filters = [
            FilenameFilter("apps.py"),
        ]
        self.app = prepare_app(jinja_app, profiler, filters=filters)

    def time_10_items(self, *args):
        self.app.get("/10")

    def time_100_items(self, *args):
        self.app.get("/20")


class ReAppTest(object):
    param_names = ["profiler"]
    params = ["base", "sync", "async"]

    def setup(self, profiler):
        self.app = prepare_app(re_app, profiler)
        chars = string.ascii_uppercase + string.ascii_lowercase + string.digits
        random.seed(12345)
        random_str = "".join(random.choice(chars) for _ in range(2000))
        self.random_str = "/" + random_str
        random.seed()

    def time_index_page(self, *args):
        self.app.get("/")

    def time_email_address(self, *args):
        self.app.get("/EMAIL-address-8324236rq@example.com")

    def time_ip_address(self, *args):
        self.app.get("/ip-address-is-123.21.13.149")

    def time_random(self, *args):
        self.app.get(self.random_str)



from types import CodeType
from typing import Any, Dict, Tuple


class LineTiming:
    code: Any
    lineno: int
    total_time: int
    n_hits: int

    def as_tuple(self) -> Tuple[int, int, int]: ...
    def __repr__(self) -> str: ...


class LastTime: ...


class LineProfiler:
    results: Dict[CodeType, Dict[int, LineTiming]]
    last_time: Dict[CodeType, LastTime]

    def enable(self) -> None: ...
    def disable(self) -> None: ...
    def reset(self) -> None: ...
    @staticmethod
    def get_unit() -> float: ...



from cpython cimport PyObject

from header cimport (
    PyEval_SetTrace, PyFrameObject, PY_LONG_LONG, PyTrace_LINE, PyTrace_RETURN
)


cdef extern from "timer.h":
    PY_LONG_LONG hpTimer()
    double hpTimerUnit()


cdef class LineProfiler:
    cdef public dict results
    cdef public dict last_time

    def __init__(self):
        self.results = {}
        self.last_time = {}

    def enable(self):
        PyEval_SetTrace(python_trace_callback, self)

    def disable(self):
        PyEval_SetTrace(NULL, <object>NULL)

    def reset(self):
        self.results = {}
        self.last_time = {}

    @staticmethod
    def get_unit():
        return hpTimerUnit()


cdef class LineTiming:
    cdef public object code
    cdef public int lineno
    cdef public PY_LONG_LONG total_time
    cdef public long n_hits

    def __init__(self, object code, int lineno):
        self.code = code
        self.lineno = lineno
        self.total_time = 0
        self.n_hits = 0

    cdef hit(self, PY_LONG_LONG dt):
            self.n_hits += 1
            self.total_time += dt

    def as_tuple(self):
        return self.lineno, self.n_hits, self.total_time

    def __repr__(self):
        return '<LineTiming for %r lineno: %r n_hits: %r total_time: %r>' % (
            self.code, self.lineno, self.n_hits, <long>self.total_time)


cdef class LastTime:
    cdef int f_lineno
    cdef PY_LONG_LONG time

    def __cinit__(self, int f_lineno, PY_LONG_LONG time):
        self.f_lineno = f_lineno
        self.time = time


cdef int python_trace_callback(object self_, PyFrameObject *py_frame, int what,
                               PyObject *arg):
    cdef LineProfiler self
    cdef dict results
    cdef dict result_code
    cdef dict last_time
    cdef LineTiming entry
    cdef LastTime old
    cdef object code
    cdef PY_LONG_LONG time
    cdef int lineno

    if what != PyTrace_LINE and what != PyTrace_RETURN:
        return 0

    time = hpTimer()

    self = <LineProfiler>self_
    last_time = self.last_time
    results = self.results
    code = <object>py_frame.f_code

    if code not in results:
        results[code] = {}

    if code in last_time:
        result_code = results[code]
        old = last_time[code]
        lineno = old.f_lineno

        if lineno not in result_code:
            result_code[lineno] = entry = LineTiming(code, lineno)
        else:
            entry = result_code[lineno]

        entry.hit(time - old.time)

        if what == PyTrace_RETURN:
            del last_time[code]

    if what == PyTrace_LINE:
        last_time[code] = LastTime(py_frame.f_lineno, hpTimer())

    return 0



from wsgi_lineprof.extensions import LineProfiler as _LineProfiler  # Cython
from wsgi_lineprof.stats import LineProfilerStat, LineProfilerStats


class LineProfiler(_LineProfiler):
    def get_stats(self):
        # type: () -> LineProfilerStats
        return LineProfilerStats(
            [LineProfilerStat(
                # TODO: Improve how to handle empty filename
                code.co_filename if code.co_filename is not None else "",
                code.co_name,
                code.co_firstlineno,
                timings)
             for code, timings in self.results.items()])



import inspect
import itertools
import linecache
from os import path
from typing import Any, Callable, Dict, Iterable, Sequence, Union  # noqa: F401

from wsgi_lineprof.extensions import (  # noqa: F401
    LineProfiler as _LineProfiler, LineTiming
)
from wsgi_lineprof.filters import BaseFilter
from wsgi_lineprof.types import Stream  # noqa: F401


class LineProfilerStat(object):
    def __init__(self,
                 filename,  # type: str
                 name,  # type: str
                 firstlineno,  # type: int
                 timings  # type: Dict[int, LineTiming]
                 ):
        # type: (...) -> None
        self.filename = filename
        self.name = name
        self.firstlineno = firstlineno
        self.timings = timings
        total_time = sum(t.total_time for t in timings.values())
        self.total_time = total_time * _LineProfiler.get_unit()

    def write_text(self, stream):
        # type: (Stream) -> None
        stream.write("File: %s\n" % self.filename)
        stream.write("Name: %s\n" % self.name)
        stream.write("Total time: %g [sec]\n" % self.total_time)
        if not path.exists(self.filename):
            # e.g., filename is <frozen importlib._bootstrap>
            stream.write("WARNING: Cannot fild a file\n")
            return

        linecache.clearcache()
        lines = linecache.getlines(self.filename)  # type: Sequence[str]
        if self.name != "<module>":
            lines = inspect.getblock(lines[self.firstlineno - 1:])

        template = '%6s %9s %12s  %-s'
        header = template % ("Line", "Hits", "Time", "Code")
        stream.write(header)
        stream.write("\n")
        stream.write("=" * len(header))
        stream.write("\n")

        d = {}  # type: Dict[int, Dict[str, Any]]
        for i, code in zip(itertools.count(self.firstlineno), lines):
            timing = self.timings.get(i)
            if timing is None:
                d[i] = {
                    "hits": "",
                    "time": "",
                    "code": code
                }
            else:
                d[i] = {
                    "hits": timing.n_hits,
                    "time": timing.total_time,
                    "code": code
                }
        for i in sorted(d.keys()):
            r = d[i]
            stream.write(template % (i, r["hits"], r["time"], r["code"]))
        stream.write("\n")


CallableFilterType = Callable[[Iterable[LineProfilerStat]],
                              Iterable[LineProfilerStat]]
FilterType = Union[CallableFilterType, BaseFilter]


class LineProfilerStats(object):
    def __init__(self, stats):
        # type: (Iterable[LineProfilerStat]) -> None
        self.stats = stats

    def write_text(self, stream):
        # type: (Stream) -> None
        stream.write("Time unit: %s [sec]\n\n" % _LineProfiler.get_unit())
        for stat in self.stats:
            stat.write_text(stream)

    def filter(self, f):
        # type: (FilterType) -> LineProfilerStats
        if isinstance(f, BaseFilter):
            return LineProfilerStats(f.filter(self.stats))
        else:
            return LineProfilerStats(f(self.stats))



from io import open
from os import path
from warnings import warn

from setuptools import Extension, find_packages, setup


root = path.abspath(path.dirname(__file__))

with open(path.join(root, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

source = "extensions/extensions."

try:
    from Cython.Build import cythonize
    source += "pyx"
except ImportError:
    def cythonize(extensions):
        return extensions
    source += "c"
    if not path.exists(path.join(root, source)):
        raise Exception("No Cython installation, no generated C file")
    warn("Could not import Cython, using generated C source code instead")

setup(
    name="wsgi_lineprof",

    version="0.6.0",

    description="WSGI middleware for line-by-line profiling",
    long_description=long_description,

    url="https://github.com/ymyzk/wsgi_lineprof",

    author="Yusuke Miyazaki",
    author_email="miyazaki.dev@gmail.com",

    license="MIT",

    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: C",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development",
    ],

    # What does your project relate to?
    # keywords="sample setuptools development",

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=["contrib", "docs", "tests"]),
    package_data={
        "wsgi_lineprof": ["py.typed", "*.pyi"],
    },

    ext_package="wsgi_lineprof",
    ext_modules=cythonize([
        Extension("extensions",
                  sources=[source, "extensions/timer.c"])
    ]),

    install_requires=[
        "six>=1.10.0",
    ],

    extras_require={
        ":python_version < '3.5'": ["typing"],
        "benchmark": [
            "asv>=0.3.1,<0.4",
            "virtualenv",
        ],
        "benchmark-deps": [
            "Jinja2==2.10",
            "WebTest==2.0.32",
        ],
        "build": ["Cython>=0.28,<0.30"],
        "docs": ["Sphinx>=1.4,<1.5"],
        "test": [
            "flake8>=3.0.0,<4.0.0",
            "pytest>=3.0.0,<4.0.0",
        ],
        "test:python_version>='3.0'": [
            "mypy>=0.521,<1.0",
        ]
    },
)



import atexit
import sys
from typing import (Any, Callable, Iterable, Optional,  # noqa: F401
                    TYPE_CHECKING)  # noqa: F401

from wsgi_lineprof.profiler import LineProfiler
from wsgi_lineprof.stats import FilterType  # noqa: F401
from wsgi_lineprof.types import Stream  # noqa: F401
from wsgi_lineprof.writers import (AsyncWriter, BaseWriter,  # noqa: F401
                                   SyncWriter)  # noqa: F401


if TYPE_CHECKING:
    from wsgiref.types import (StartResponse, WSGIApplication,  # noqa: F401
                               WSGIEnvironment)  # noqa: F401


class LineProfilerMiddleware(object):
    def __init__(self,
                 app,  # type: WSGIApplication
                 stream=None,  # type: Optional[Stream]
                 filters=tuple(),  # type: Iterable[FilterType]
                 async_stream=False,  # type: bool
                 accumulate=False,  # type: bool
                 ):
        # type: (...) -> None
        self.app = app
        self.stream = sys.stdout if stream is None else stream  # type: Stream
        self.filters = filters
        self.accumulate = accumulate
        self.profiler = LineProfiler()
        # Cannot use AsyncWriter with atexit
        if async_stream and not accumulate:
            self.writer = AsyncWriter(self.stream)  # type: BaseWriter
        else:
            self.writer = SyncWriter(self.stream)  # type: BaseWriter
        if accumulate:
            atexit.register(self._write_stats)

    def _write_stats(self):
        # type: () -> None
        stats = self.profiler.get_stats()
        for f in self.filters:
            stats = stats.filter(f)
        self.writer.write(stats)

    def __call__(self, env, start_response):
        # type: (WSGIEnvironment, StartResponse) -> Iterable[bytes]
        if not self.accumulate:
            self.profiler.reset()
        self.profiler.enable()
        result = self.app(env, start_response)
        self.profiler.disable()

        if not self.accumulate:
            self._write_stats()

        return result



from abc import ABCMeta, abstractmethod
from six import add_metaclass
from six.moves import queue
from threading import Thread
from typing import Any  # noqa: F401

from wsgi_lineprof.stats import LineProfilerStats  # noqa: F401
from wsgi_lineprof.types import Stream  # noqa: F401


@add_metaclass(ABCMeta)
class BaseWriter(object):
    @abstractmethod
    def __init__(self,
                 stream,  # type: Stream
                 *kwargs  # type: Any
                 ):
        # type: (...) -> None
        return

    @abstractmethod
    def write(self, stats):
        # type: (LineProfilerStats) -> None
        return


class SyncWriter(BaseWriter):
    def __init__(self,
                 stream,  # type: Stream
                 ):
        # type: (...) -> None
        self.stream = stream

    def write(self, stats):
        # type: (LineProfilerStats) -> None
        stats.write_text(self.stream)


class AsyncWriter(BaseWriter):
    def __init__(self,
                 stream,  # type: Stream
                 ):
        # type: (...) -> None
        self.stream = stream
        self.queue = queue.Queue()  # type: queue.Queue
        self.writer_thread = Thread(target=self._write)
        self.writer_thread.setDaemon(True)
        self.writer_thread.start()

    def write(self, stats):
        # type: (LineProfilerStats) -> None
        self.queue.put(stats)

    def _write(self):
        # type: () -> None
        while True:
            stats = self.queue.get()
            stats.write_text(self.stream)
